/*
 * Copyright 2023 Denis Pynkin <denis.pynkin@collabora.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuQcS5gen2Plugin, fu_qc_s5gen2_plugin, FU, AUDIO_S5GEN2_PLUGIN, FuPlugin)
