/*
 * Copyright 2024 Algoltek, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuAlgoltekUsbcrPlugin,
		     fu_algoltek_usbcr_plugin,
		     FU,
		     ALGOLTEK_USBCR_PLUGIN,
		     FuPlugin)
