// Copyright 2025 Richard Hughes <richard@hughsie.com>
// SPDX-License-Identifier: LGPL-2.1-or-later

enum FuPciDeviceBaseCls {
    Old,
    MassStorage,
    Network,
    Display,
    Multimedia,
    Memory,
    Bridge,
    SimpleCommunication,
    Base,
    Input,
    Docking,
    Processors,
    SerialBus,
    Wireless,
    IntelligentIo,
    Satellite,
    Encryption,
    SignalProcessing,
    Accelerator,
    NonEssential,
    Undefined = 0xFF,
}
