/*
 * Copyright 2024 Mike Chang <Mike.chang@telink-semi.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuTelinkDfuPlugin, fu_telink_dfu_plugin, FU, TELINK_DFU_PLUGIN, FuPlugin)
