/*
 * Copyright 2024 HP Development Company, L.P.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuHpiCfuPlugin, fu_hpi_cfu_plugin, FU, HPI_CFU_PLUGIN, FuPlugin)
