/*
 * Copyright 2025 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#define FU_ELAN_KBD_DEVICE_ADDR_BOOT   0x0
#define FU_ELAN_KBD_DEVICE_ADDR_APP    0x2000
#define FU_ELAN_KBD_DEVICE_ADDR_OPTION 0x8000

#define FU_ELAN_KBD_DEVICE_SIZE_ROM 0x8000

#define FU_ELAN_KBD_DEVICE_SIZE_BOOT   (FU_ELAN_KBD_DEVICE_ADDR_APP - FU_ELAN_KBD_DEVICE_ADDR_BOOT)
#define FU_ELAN_KBD_DEVICE_SIZE_APP    (FU_ELAN_KBD_DEVICE_SIZE_ROM - FU_ELAN_KBD_DEVICE_ADDR_APP)
#define FU_ELAN_KBD_DEVICE_SIZE_OPTION 0x100
